import re

from typing import List, Tuple, Optional

from mypy import api as mypy_api

from svst.config import read_configuration


config = read_configuration()


def is_ignored_file(root: str, file: str) -> bool:
    """Return if the file is ignored by the configuration.

    Args:
        root: Directory path str.
        file: Name of the file.

    Returns:
        If the file needs to be ignored based on the configurations.
    """

    if not file.endswith(".py"):
        return True

    if not config:
        return False

    for prefix_path in (
        [item for item in config["ignore_path_prefix"].split("\n") if item]
        if "ignore_path_prefix" in config
        else []
    ):
        if root.replace("./", "", 1).startswith(prefix_path):
            return True

    for directory in (
        [item for item in config["ignore_directory"].split("\n") if item]
        if "ignore_directory" in config
        else []
    ):
        if directory in root:
            return True

    if (
        file in [item for item in config["ignored_file_names"].split("\n") if item]
        if "ignored_file_names" in config
        else []
    ):
        return True

    return False


def run_and_clean_mypy_output_in_path(
    path: str,
) -> List[str]:
    """Use mypy api to analyse a directory or file.

    Args:
        path: file or directory path str.

    Returns:
        str:
    """
    results: Tuple[str, str, int] = mypy_api.run([path])

    for error_message in results[0].split("\n")[:-2]:
        yield error_message



import os
import configparser

from svst import constants

from typing import Optional, Dict


def _get_configuration_path() -> Optional[str]:
    """Get configuration file path from project root directory."""

    root_directory: str = os.getcwd()
    if root_directory:
        file_name: str
        for file_name in constants.POSSIBLE_CONFIGURATION_FILES:
            file_path: str = os.path.join(root_directory, file_name)
            if os.path.exists(file_path):
                return file_path

    return None


def read_configuration() -> Optional[Dict[str, str]]:
    """Get configuration keys from file.

    Returns:
        When a configuration file is available the config Dict is returned.
    """

    config_path: str = _get_configuration_path()

    if not config_path:
        return None

    config: configparser.ConfigParser = configparser.ConfigParser()
    config.read(config_path)

    return config["svst"] if "svst" in config else None



import ast

from typing import List, Dict, Union, Optional

from svst import constants, output


class ParentNodeTransformer(ast.NodeTransformer):
    def visit(self, node):
        child: ast.NodeVisitor
        for child in ast.iter_child_nodes(node):
            child.parent = node
        return super().visit(node)


class StaticTypeEnforcer(ast.NodeVisitor):
    def __init__(
        self,
        file_name: Optional[str] = None,
        logging_level: str = constants.STANDARD_LOGGING_LEVEL,
    ) -> None:
        self.scope: str = "global"
        self.assignments: Dict[str, set] = {"global": set()}
        self.type_annotations: Dict[str, set] = {"global": set()}

        self.file_name = file_name
        self.logging_level = logging_level

        self.output: List[Dict[str, Union[str, int]]] = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        previous_scope = self.scope
        self.scope = node.name
        self.assignments[self.scope] = set()
        self.type_annotations[self.scope] = set()
        self.generic_visit(node)
        self.scope = previous_scope

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_FunctionDef(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if isinstance(node.target, ast.Name):
            var_name = node.target.id
            self.type_annotations[self.scope].add(var_name)
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            var_name = node.targets[0].id
            if var_name not in self.type_annotations[self.scope]:
                if var_name not in self.assignments[self.scope]:
                    self.output.append(
                        output.output_structure_constructor(
                            self.file_name,
                            node.lineno,
                            var_name,
                            self.scope,
                            self.logging_level,
                        )
                    )
                self.assignments[self.scope].add(var_name)
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        if isinstance(node.target, ast.Name):
            var_name = node.target.id
            if var_name not in self.type_annotations[self.scope]:
                self.output.append(
                    output.output_structure_constructor(
                        self.file_name,
                        node.lineno,
                        var_name,
                        self.scope,
                        self.logging_level,
                    )
                )
        if isinstance(node.target, ast.Tuple):
            for elt in node.target.elts:
                if isinstance(elt, ast.Name):
                    var_name = elt.id
                    if var_name not in self.type_annotations[self.scope]:
                        self.output.append(
                            output.output_structure_constructor(
                                self.file_name,
                                node.lineno,
                                var_name,
                                self.scope,
                                self.logging_level,
                            )
                        )

        self.generic_visit(node)



from typing import Optional
from typing_extensions import TypedDict

from svst import constants


class OutputTypedDict(TypedDict):
    file_name: Optional[str]
    line_number: int
    variable_name: str
    variable_scope: str
    logging_level: str


def output_structure_constructor(
    file_name: Optional[str],
    line_number: int,
    variable_name: str,
    variable_scope: str,
    logging_level: str = constants.STANDARD_LOGGING_LEVEL,
) -> OutputTypedDict:
    """Constructor for dictionary output.

    Args:
        file_name: Name of the file being analysed. None when parse_code is used on bare str.
        line_number: int number of the line where the error occurred.
        variable_name: Name of the variable where the error occurred.
        variable_scope: Scope of the variable where the error occurred.
        logging_level: str of logging level configuration.

    Returns:
        A Dict that contains the info of the svst errors:

       {
           "file_name": "some_file.py",
           "line_number": 6,
           "variable_name": "yet_another_variable",
           "variable_scope": "some_method",
           "logging_level": "DEBUG",
       }

       This dictionary is the first layer of svst errors construction.
    """

    return {
        "file_name": file_name,
        "line_number": line_number,
        "variable_name": variable_name,
        "variable_scope": variable_scope,
        "logging_level": logging_level,
    }


def output_string_constructor(
    output_structure: OutputTypedDict,
) -> str:
    """Constructor for string output.

    Args:
        output_structure: {
           "file_name": "some_file.py",
           "line_number": 6,
           "variable_name": "yet_another_variable",
           "variable_scope": "some_method",
           "logging_level": "DEBUG",
       }

    Returns:
        str: {file_name}:{line_number}: error: Variable "{variable_name}" is missing a
        standalone variable type annotation in the scope "{variable_scope}"  [no-untyped-var]
    """

    return (
        f"{output_structure['file_name']}:{output_structure['line_number']}: "
        f"error: Variable \"{output_structure['variable_name']}\" is missing "
        f"a standalone variable type annotation in the "
        f"scope \"{output_structure['variable_scope']}\"  [no-untyped-var]"
    )



# Always prefer setuptools over distutils
from setuptools import setup  # type: ignore

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE: str = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description: str = f.read()

# This call to setup() does all the work
setup(
    name="svst",
    version="0.1.6",
    description="Standalone Variable Static Typing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://svst.readthedocs.io/",
    author="Spiked Engineering",
    author_email="spikedengineering@tutanota.com",
    project_urls={"Code Repository": "https://github.com/spikedengineering/svst"},
    license="MPL2",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    packages=["svst"],
    include_package_data=True,
    install_requires=["mypy"],
    entry_points={
        "console_scripts": [
            "svst = svst.cli:main",
        ],
    },
)



import sys

from svst import api

from typing import List

import argparse


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--mypy", action="store_true")
    parser.add_argument("--check", action="store_true")
    parser.add_argument("paths", nargs="+")

    arguments = parser.parse_args()

    errors_count: int = 0
    error_messages: List[str] = api.run(arguments.paths, mypy=arguments.mypy)
    for error_line in error_messages:
        errors_count += 1
        if not arguments.check:
            print(error_line, file=sys.stderr)

    summary_message = (
        f"Found {errors_count} errors running svst and mypy."
        if arguments.mypy
        else f"Found {errors_count} errors running svst."
    )

    if errors_count:
        print(summary_message, file=sys.stderr)
        sys.exit(1)
    else:
        print(summary_message)



import ast
import os
import io

from svst import utils, output, parsing

from typing import List, Optional

from svst import constants


def parse_code(
    code: str,
    logging_level: str = constants.STANDARD_LOGGING_LEVEL,
    file_name: Optional[str] = None,
) -> List[output.OutputTypedDict]:
    """Parse code directly and return a list of output dictionaries.

    Args:
        code: str of python code; can have multiple lines ofc.
        logging_level: str of logging level configuration.
        file_name: Path of the file being analysed.

    Returns:
        A List of Dict that contains the info of the svst errors:

       [{ "file_name": "some_file.py",
          "line_number": 6,
          "variable_name": "yet_another_variable",
          "variable_scope": "some_method",
          "logging_level": "DEBUG" }, ... ]
    """

    tree = ast.parse(code)
    ast.increment_lineno(tree)
    ast.fix_missing_locations(tree)
    parsing.ParentNodeTransformer().visit(tree)
    visitor = parsing.StaticTypeEnforcer(file_name, logging_level)
    visitor.visit(tree)

    return visitor.output


def run(
    path_list: List[str],
    logging_level: str = constants.STANDARD_LOGGING_LEVEL,
    mypy: bool = False,
) -> List[str]:
    """Run path list.

    Args:
        path_list: List of str paths.
        logging_level: str of logging level configuration.
        mypy: bool for running mypy alongside svst.

    Returns:
        List of error str: ['{file_name}:{line_number}: error: Variable "{variable_name}" is missing a
        standalone variable type annotation in the scope "{variable_scope}"  [no-untyped-var]', ...]
    """

    path: str
    for path in path_list:
        root: str
        dirs: List[str]
        files: List[str]
        for root, dirs, files in os.walk(path):
            file: str
            for file in files:
                if utils.is_ignored_file(root, file):
                    continue

                file_path: str = os.path.join(root, file)

                file_buffer: io.TextIOWrapper
                with open(file_path, "r") as file_buffer:
                    svst_errors = parse_code(
                        file_buffer.read(), logging_level, file_path
                    )

                    for svst_error in svst_errors:
                        yield output.output_string_constructor(svst_error)

        if mypy:
            for line in utils.run_and_clean_mypy_output_in_path(path):
                yield line
